from django.apps import apps
from django.contrib.auth.models import User
from django.template import Context, Template
from django.test import TestCase

from django_policies.registry import default_registry
from testapp.models import Article


class DjangoIntegrationTest(TestCase):
    def setUp(self):
        with self.settings(POLICIES_AUTODISCOVER=True):
            aconfig = apps.get_app_config("django_policies")
            aconfig.ready()

    def test_autodiscover(self):
        self.assertTrue(len(default_registry._registry) > 0)

    def test_permission_check(self):
        u = User.objects.create_user("admin", "admin@localhost", "admin")

        self.assertTrue(u.has_perm("testapp.add_article"))
        self.assertTrue(u.has_perm("testapp.change_article", Article()))
        self.assertFalse(u.has_perm("testapp.delete_article"))

    def test_template_tag(self):
        user = User.objects.create_user("admin", "admin@localhost", "admin")

        tests = [
            ('user "testapp.add_article"', "True"),
            ('user "testapp.change_article" article', "True"),
            ('user "testapp.change_article"', "False"),
            ('user "testapp.delete_article"', "False"),
            ('user "random"', "False"),
        ]

        for test in tests:
            template = Template(
                "{% load policies %}{% has_perm " + test[0] + " as p %}{{ p }}"
            )
            ctx = Context({"user": user, "article": Article()})
            result = template.render(ctx)
            self.assertEqual(result, test[1])



from django.apps import AppConfig
from django.conf import settings

from django_policies.registry import autodiscover_policies


class DjangoPoliciesConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "django_policies"

    def ready(self):
        if not getattr(settings, "POLICIES_AUTODISCOVER", True):
            return

        autodiscover_policies(
            getattr(settings, "POLICIES_AUTODISCOVER_MODULE", "policies")
        )



from django.contrib.auth.models import AnonymousUser, User
from django.test import TestCase

from django_policies.policies import Policy


class PolicyTest(TestCase):
    def test_get_permission_methods(self):
        tests = [
            ("testapp.add_article", {"can_add", "can_add_article"}),
            ("testapp.publish_article", {"can_publish", "can_publish_article"}),
            (
                "testapp.do_something_article",
                {"can_do_something", "can_do_something_article"},
            ),
            ("testapp.nounderscores", {"can_nounderscores"}),
            ("noapp", {"can_noapp"}),
            ("noapp_article", {"can_noapp", "can_noapp_article"}),
        ]

        p = Policy()
        for inp, out in tests:
            self.assertEqual(p.get_permission_methods(inp), out)

    def test_check_permission_object(self):
        class TestPolicy(Policy):
            def can_add(self, *args):
                self.add_called = args
                return True

            def can_add_some(self, *args):
                self.add_some_called = args
                return True

            def can_publish_article(self, *args):
                self.publish_called = args
                return True

            can_change = True

        user = User()
        obj = 34

        p = TestPolicy()
        res = p.check_permission(user, "testapp.add_article", obj)
        self.assertTrue(res)
        self.assertTrue(hasattr(p, "add_called"))
        self.assertFalse(hasattr(p, "add_some_called"))
        self.assertFalse(hasattr(p, "publish_called"))
        self.assertEqual(p.add_called, (user, obj))

        p = TestPolicy()
        res = p.check_permission(user, "testapp.publish_article", obj)
        self.assertTrue(res)
        self.assertTrue(hasattr(p, "publish_called"))
        self.assertFalse(hasattr(p, "add_some_called"))
        self.assertFalse(hasattr(p, "add_called"))
        self.assertEqual(p.publish_called, (user, obj))

        res = p.check_permission(user, "some_random_permission", obj)
        self.assertFalse(res)

        res = p.check_permission(user, "testapp.change_article", obj)
        self.assertTrue(res)

    def test_check_permission_no_object(self):
        class TestPolicy(Policy):
            def can_add(self, *args):
                self.add_called = args
                return True

            def can_add_some(self, *args):
                self.add_some_called = args
                return True

            can_change_some = True

        user = User()

        p = TestPolicy()
        res = p.check_permission(user, "testapp.add_article")
        self.assertTrue(res)
        self.assertFalse(hasattr(p, "add_called"))
        self.assertEqual(p.add_some_called, (user,))

        res = p.check_permission(user, "testapp.change_article")
        self.assertTrue(res)

    def test_check_permission_invalid(self):
        class TestPolicy(Policy):
            def can_add_some(self, *args):
                return 12

            can_change_some = 34

        user = User()

        p = TestPolicy()
        with self.assertRaises(ValueError):
            p.check_permission(user, "testapp.add_article")

        with self.assertRaises(ValueError):
            p.check_permission(user, "testapp.change_article")

    def test_anonymous(self):
        class TestPolicy(Policy):
            can_change = True

        tests = [
            (True, False),
            (False, True),
        ]

        auser = AnonymousUser()

        for anonymous, expected in tests:
            TestPolicy.deny_anonymous = anonymous
            self.assertEqual(
                TestPolicy().check_permission(auser, "testapp.change_article", 12),
                expected,
            )



from typing import Any
from django.contrib.auth.mixins import (
    PermissionRequiredMixin as BasePermissionRequiredMixin,
)
from django.views.generic.edit import BaseCreateView


class PermissionRequiredMixin(BasePermissionRequiredMixin):
    def get_permission_object(self) -> Any:
        if hasattr(self, "object"):
            return getattr(self, "object")

        if not isinstance(self, BaseCreateView):
            if hasattr(self, "get_object"):
                return getattr(self, "get_object")()

        return None

    def has_permission(self) -> bool:
        obj = self.get_permission_object()
        perms = self.get_permission_required()
        return self.request.user.has_perms(perms, obj)



from typing import TYPE_CHECKING, Any

from django.conf import settings

if TYPE_CHECKING:  # pragma: no cover
    from django.contrib.auth.models import AbstractUser, AnonymousUser

    AnyUser = AbstractUser | AnonymousUser


class Policy:
    no_object_suffix: str = "_some"
    deny_anonymous: bool | None = None

    def get_permission_methods(self, permission: str) -> set[str]:
        methods = set()
        permission = permission.split(".", 1)[-1]

        # can_change_model
        methods.add(f"can_{permission}")

        # can_change
        perm_name = permission.rsplit("_", 1)[0]
        methods.add(f"can_{perm_name}")

        return methods

    def check_permission(
        self, user: "AnyUser", permission: str, obj: Any = None
    ) -> bool:
        if self.deny_anonymous is None:
            self.deny_anonymous = getattr(
                settings, "POLICIES_DENY_ANONYMOUS_DEFAULT", False
            )

        for method in self.get_permission_methods(permission):
            if obj is None:
                method += self.no_object_suffix

            if not hasattr(self, method):
                continue

            if self.deny_anonymous and not user.is_authenticated:
                return False

            attr = getattr(self, method)
            if callable(attr):
                args = [user]
                if obj is not None:
                    args.append(obj)

                val = attr(*args)
                if not isinstance(val, bool):
                    raise ValueError(f"{method} returned {type(val)}, expected bool.")
                return val
            if isinstance(attr, bool):
                return attr

            raise ValueError(
                f"Invalid value of {method} in {self.__class__.__name__}: expected boolean or a callable."
            )

        return False



from django.test import TestCase

from django_policies.decorators import register
from django_policies.policies import Policy
from django_policies.registry import PolicyRegistry, default_registry
from testapp.models import Article


class RegisterDecoratorTest(TestCase):
    def test_register(self):
        r = PolicyRegistry()

        @register(Article, r)
        class TestPolicy(Policy):
            pass

        self.assertIn("testapp.add_article", r._registry)

    def test_register_non_policy(self):
        r = PolicyRegistry()

        with self.assertRaises(ValueError):

            @register(Article, r)
            class FakePolicy:
                pass

        self.assertNotIn("testapp.add_article", r._registry)

    def test_default(self):
        @register(Article)
        class TestPolicy(Policy):
            pass

        self.assertIn("testapp.add_article", default_registry._registry)
        default_registry.unregister(Article)
